/**********************************************************
** 
**  HMCP EXTENSION  1  algorithm 
**
**  author: Haluk   
**  rev 1.0 4/1/98
**  
**********************************************************/




#include <stdio.h>
#include <time.h>

#define ALGORITHM seq

#define MAXNUM          (1.0*0x7FFFFFFF)
#define MAX(x,y)        ((x == MAXNUM || x > y) ? x : y)
#define ABS(x)          (((x) < 0.0) ? (-1.0 * (x)) : (x))
#define MIN(x,y)        ((x == MAXNUM || x > y) ? y : x)
#define GREATER(x,y)    ((x == MAXNUM || x > y) ? TRUE : FALSE)
#define MID_ITEM(x)      (((x % 2) == 0) ? x/2 : (x+1)/2)
#define MAXTPE          64
#define MAXNODE         4096
#define ERROR_LIMIT     1.0e-5
#define FILE_NAME       "TEST_FILE"

#define DONE	1
#define HALT	2

#define TRUE    1
#define FALSE   0

typedef struct str_link { /* child/parent of a graph node */ 
    int    nid; 
    float   commcost;
    struct str_link *linkNext; 
} LINK; 
 
typedef struct str_node { /* a graph node */ 
    float      compcost; 
    float      asap, alap;
    float      tlevel, blevel;
    int        childnum;
    int        visit, visit_p, pnum;
    int        nid,i;
    LINK       *child, *parent;
    float      start, end;
    int        schPe, scheduled;
    int        pseudo_p, pseudo_c;
    struct str_node *schNext, *schLast;
} NODE; 

float  comp_weights[MAXNODE][MAXTPE], value[MAXTPE];
int    place[MAXTPE];

void b_level();
void t_level();
void insert_list(int pos);
int  get_highest_p_ready_node();
void QuickSort(int low, int high);
float median(int nid);
void median_find(int low, int high, int mid);

NODE *graph, *schHead[MAXTPE], *schTail[MAXTPE];
int   *pri_list;

int  *blevel_order, *ready, *ready2, readynum, readynum2, node_id;
int  list_length = 0;

int  numnode = 0; /* graph node information */
int  numtpe = 4;  /* target PE information */
int  numppe, pid; /* phisical PE information */
int  mid_point;   /* the processor no (which is the median for exec times */
int  nid, pe, num_pe_used, tpe, parentid, pid,  childnode, parent_pe;
float  cp_length, available, best_available, start_time, sch_length, mean_val, 
       min_val, max_val, t, cp_length_no_comm, sum_min, sum_max, total;
NODE   *nodep, *pnodep, *prev, *sch;
LINK   *ptr, *parentptr, *childptr;
FILE   *fp, *err, *res, *finals;
   
/**********
*
*    main program - 
*
**********/
void main(int argc, char **argv)
{
  int i, j, k, s, b, x, y, m, node;
  
  long b_Time, m_Time, e_Time, diff_Time;
  double  clocks_per_second = (double) CLOCKS_PER_SEC;
  float  actual_range, actual_CCR;

    if (argc < 2) {
        fprintf(stderr, "usage: %s TargetPeNum Graph\n",argv[0]);
        exit(0);
        }
    sscanf(argv[1],"%d",&numtpe);

    if ((fp = fopen(FILE_NAME, "r")) == NULL) {
        fprintf(stderr, "Error in reading graph file %s\n",argv[2]);
        exit(0);
        }

    fscanf(fp, "%d %f %f", &numnode, &actual_CCR, &actual_range); 
     /* read number of nodes, CCR, range */ 

    if ((res = fopen("m2.out", "a")) == NULL) {
        fprintf(stderr, "Error in writing graph file \n");
        exit(0);
        }
    if ((pri_list = (int*) malloc ((numnode+10)*sizeof(int))) == NULL) { 
        fprintf(stderr, "Error in allocating memory, %d nodes\n",numnode);
        exit(0);
        }
    if ((graph = (NODE*) malloc ((numnode+1)*sizeof(NODE))) == NULL) { 
        fprintf(stderr, "Error in allocating memory, %d nodes\n",numnode);
        exit(0);
        }
    if ((blevel_order = (int*) malloc (numnode*sizeof(int))) == NULL) { 
        fprintf(stderr, "Error in allocating memory, %d nodes\n",numnode);
        exit(0);
        }
    if ((ready = (int*) malloc (numnode*sizeof(int))) == NULL) { 
        fprintf(stderr, "Error in allocating memory, %d nodes\n",numnode);
        exit(0);
        }
    if ((ready2 = (int*) malloc (numnode*sizeof(int))) == NULL) { 
        fprintf(stderr, "Error in allocating memory, %d nodes\n",numnode);
        exit(0);
        } 
    
    for (i = readynum = readynum2 = 0; i < numnode; i++) { 
        /* initialization */
	graph[i].nid = i;
        graph[i].schPe = -1;
        graph[i].pnum  = 0;
	graph[i].child = graph[i].parent = NULL;
        graph[i].start = graph[i].end = graph[i].tlevel = graph[i].blevel= 0.0;
	graph[i].asap = graph[i].alap = 0.0;
	graph[i].visit = graph[i].scheduled = 0;
        fscanf(fp, "%d", &nid); /* get node id */ nid == i;
        /* then read comp cost and number of children */
        fscanf(fp, "%d", &(graph[i].childnum));
        for (j = 0; j < graph[i].childnum; j++) { 
            if ((ptr = (LINK*) malloc (sizeof(LINK))) == NULL) { 
                fprintf(stderr, "ERROR : memory for edge of node %d\n", i+1); 
                exit(0); 
                } 
            fscanf(fp, "%d%f", &(ptr->nid), &(ptr->commcost));
            ptr->nid--; 
            ptr->linkNext = graph[i].child; graph[i].child = ptr; 
	} 
        if (graph[i].child==NULL) 
	     ready[readynum++] = i;
    }

    /* Build up parent list for each node */ 
    for (i = 0; i < numnode; i++) {
        for (ptr=graph[i].child; ptr!=NULL; ptr=ptr->linkNext) { 

            childnode = ptr->nid; 
            if ((parentptr = (LINK*) malloc (sizeof(LINK))) == NULL) { 
                fprintf(stderr, "ERROR : memory for parent node of %d\n", i+1); 
                exit(0); 
                } 
            parentptr->nid = i; 
            parentptr->commcost = ptr->commcost; 
            parentptr->linkNext = graph[childnode].parent;
            graph[childnode].parent = parentptr;
            graph[childnode].pnum++;
	}
	if (graph[i].parent==NULL)  
	    ready2[readynum2++] =i;
    } 
    
    /* reading the computation costs from the file */
    for (m = 0; m < numnode; m++) {
      fscanf(fp, "%d", &node_id);
      for (j = 0; j <numtpe; j++) 
           fscanf(fp, "%f", &comp_weights[m][j]);
    }


#include  "graph.att"

#ifdef MEAN_V
     /* computation costs for mean */
    if ((finals = fopen("m2.mean", "a")) == NULL) {
        fprintf(stderr, "Error in appending the file");
        exit(0);
    }
    for (m = 0; m < numnode; m++) {
       for (mean_val=0.0, j = 0; j <numtpe; j++) {
	   mean_val+= comp_weights[m][j];
       }
       graph[m].compcost= mean_val/numtpe;
    }
#endif 

#ifdef MIN_V
    /* computation costs for min */
    if ((finals = fopen("m2.min", "a")) == NULL) {
        fprintf(stderr, "Error in appending the file");
        exit(0);
    }
    for (m = 0; m < numnode; m++) {
       for (min_val=MAXNUM, j = 0; j <numtpe; j++) {
	   if (comp_weights[m][j] < min_val)
	       min_val = comp_weights[m][j];
       }
       graph[m].compcost= min_val;
    }
#endif 

#ifdef MAX_V
    /* computation costs for min */
    if ((finals = fopen("m2.max", "a")) == NULL) {
        fprintf(stderr, "Error in appending the file");
        exit(0);
    }
    for (m = 0; m < numnode; m++) {
       for (max_val=0, j = 0; j <numtpe; j++) {
	   if (comp_weights[m][j] > max_val)
	       max_val = comp_weights[m][j];
       }
       graph[m].compcost= max_val;
    }
#endif 

#ifdef MEDIAN_V
    /* median of the computation costs */
    if ((finals = fopen("m2.median", "a")) == NULL) {
        fprintf(stderr, "Error in appending the file");
        exit(0);
    }
    for (m = 0; m < numnode; m++) {
	for (mean_val=0.0, j = 0; j <numtpe; j++)
	     graph[m].compcost = median(m);
        /* printf("for:%d median=%f \n", m, graph[m].compcost);*/
    }
#endif

    /* Done with init I/O phase */
    fclose(fp);

    b_Time = clock();

    /* find out TLEVEL time for each node, O(n+e) */
    t_level(); 

    /* find out BLEVEL time for each node, O(n+e) */
    b_level();
    cp_length = graph[0].blevel;

    for (pe = 0; pe < numtpe; pe++) { 
	schHead[pe] = schTail[pe] = NULL;
    }

    /* The root (assumed only one (for this command) is added to the heap */
    insert_list(graph[0].nid);

    /*************************************************/
    /* main algorithm: core of the heterogeneous MCP */
    for (j=0; j<numnode; j++) {
      best_available=MAXNUM;
      nid = get_highest_p_ready_node();
      for (pe=0, tpe=0; pe <numtpe; pe++) {
	for (available=0,ptr=graph[nid].parent;
	     ptr!=NULL;ptr=ptr->linkNext){
	  parentid = ptr->nid; parent_pe = graph[parentid].schPe;
	  t=graph[parentid].end;
	  if (parent_pe!=pe) t += ptr->commcost;
	  if (available < t) {available=t;}
	}
	for (prev=NULL,nodep=schHead[pe]; nodep!=NULL; 
	     prev=nodep,available=MAX(available,nodep->end),
	       nodep=nodep->schNext) 
	  if ((nodep->start-available)>=comp_weights[nid][pe]) break;
	if (available+comp_weights[nid][pe] < best_available) {
	  tpe=pe;
	  best_available=available+comp_weights[nid][pe];
	  start_time = available;
	}
      }

      graph[nid].schPe = tpe; 
      graph[nid].start = start_time;
      graph[nid].scheduled = TRUE;
      graph[nid].compcost= comp_weights[nid][tpe];
      graph[nid].end = best_available;
      /* normalde asagisi commented == 2 satir       
      fprintf(res, "nid=%d weight=%f start=%f end=%f PE=%d\n", nid,
	     comp_weights[nid][tpe], graph[nid].start, graph[nid].end, tpe); */

      for (prev=NULL,sch=schHead[tpe]; sch!=NULL; 
	   prev=sch,start_time=MAX(start_time,sch->end),
	     sch=sch->schNext) 
	if ((sch->start-start_time) >= comp_weights[nid][tpe]) break;
      
      if (prev==NULL) {
	graph[nid].schNext = schHead[tpe];
	schHead[tpe] = &(graph[nid]);
      }
      else {
	graph[nid].schNext = prev->schNext;
	prev->schNext = &(graph[nid]);
	/* graph[nid].schLast = prev; */  /*HT: updated */
      }
      /* if (graph[nid].schNext == NULL)
	 schTail[tpe] = &(graph[nid]); */ /* HT: updated */
      for  (ptr = graph[nid].child; ptr!=NULL; ptr=ptr->linkNext) {
	  pnodep = &(graph[ptr->nid]); 
	  if (++pnodep->visit_p==pnodep->pnum)
	      insert_list(pnodep->nid);
      }
    }
    /*******  end of the main part   *********************/

    e_Time = clock();
    diff_Time = e_Time - b_Time;
    sch_length=graph[blevel_order[numnode-1]].end;

    /******************  error  check & save schedule *********************/
    if ((err = fopen("m2.err", "a")) == NULL) {
        fprintf(stderr, "Error in writing graph file \n");
        exit(0);
        }
    num_pe_used = 0;
    for (sch_length=pe=i=0; pe<numtpe; i=0,pe++) {
      if (schHead[pe]!=NULL) 
	num_pe_used++;
      for (sch=schHead[pe]; sch!=NULL; sch=sch->schNext) {
	if (sch->end > sch_length) sch_length=sch->end;
	
	if (ABS(sch->end-(sch->start+graph[sch->nid].compcost)) /
	       graph[sch->nid].compcost > ERROR_LIMIT)
	 fprintf(err,"\n At PE%d node %d start %f %f end %f cost %f WRONG 1\n",
		  pe,sch->nid,sch->start,graph[sch->nid].start,sch->end,
		  graph[sch->nid].compcost);
	  /* else  fprintf(err,"%d-1 \t",sch->nid);  */

	if (sch->schNext!=NULL&&sch->end>sch->schNext->start)
	  if (((sch->end - sch->schNext->start) / sch->end) > ERROR_LIMIT)
	    fprintf(err,"\n At PE%d node %d start %f %f end %f WRONG 2\n",
		   pe,sch->nid,sch->start,graph[sch->nid].start,sch->end);
	  /* else fprintf(err, "%d-2 \t",sch->nid); */

	if (sch->start!=graph[sch->nid].start)
	  fprintf(err, "At PE%d node %d start %f %f end %f WRONG 3\n",
		 pe,sch->nid,sch->start,graph[sch->nid].start,sch->end);
      }
    }

    fprintf(finals, "%f  %f  %f  %f  %f  %ld  %d\n", sum_min, sum_max, 
	    cp_length_no_comm, cp_length, sch_length, diff_Time, num_pe_used);

    /*   writing the schedule to a file   
    for (pe=0; pe < numtpe; pe++) {
      fprintf(res,"PE=%d\t: ", pe);
      for (m=0, sch=schHead[pe]; sch!=NULL; sch=sch->schNext) {
	fprintf(res,"%d:%7.2f  ", sch->nid, sch->end);
	if ((++m) % 5 == 0) fprintf(res,"\n");
      }
      fprintf(res,"\n");
    }
    fprintf(res,"\n");    end writing */
    fclose(err);
    fclose(res);
    fclose(finals);
}

void QuickSort(int low, int high)
{   int i,j;
float k,t;
if (low == high-1) {
    if (graph[blevel_order[low]].blevel < graph[blevel_order[high]].blevel)
	{k=blevel_order[low]; blevel_order[low]=blevel_order[high];
	blevel_order[high]=k;}
    return;
}
t = graph[blevel_order[low]].blevel;
/*
  t = graph[blevel_order[j=(low+high)/2]].alap;
  k=blevel_order[low]; blevel_order[low]=blevel_order[j];blevel_order[j]=k;
  */
    for (i=low+1,j=high; i<=j;) {
        for (; i<=j && graph[blevel_order[i]].blevel>=t; i++) ;
	for (; i<=j && graph[blevel_order[j]].blevel<t; j--) ;
	if (i<j) {
	    k=blevel_order[i]; blevel_order[i]=blevel_order[j];
	    blevel_order[j]=k;
	    i++; j--;}
	}
    k=blevel_order[low]; blevel_order[low]=blevel_order[j];blevel_order[j]=k;
    if (low < j-1) QuickSort(low,j-1);
    if (j+1 < high) QuickSort(j+1,high);
}
 
float median(int nid) {
   int mid_place, i;
   for (i=0; i<numtpe; i++)  {
       value[i] = comp_weights[nid][i];
       place[i] = i;
   }
   mid_place = MID_ITEM(numtpe)-1;
   median_find(0, numtpe-1, mid_place);
   return(value[mid_point]);
}

void median_find(int low, int high, int mid) {
   int i,j,k;
   float temp, t;
   t=value[low];
   for (i=low+1,j=high; i<=j;) {
      for (; i<=j && value[i]<t; i++) ;
      for (; i<=j && value[j]>=t; j--) ;
      if (i<j) {
	 temp=value[i]; value[i]=value[j]; value[j]=temp;
	 k=place[i]; place[i]=place[j];  place[j]=k;
	 i++; j--;}
   }
   temp=value[low]; value[low]=value[j];  value[j]=temp;
   k=place[low]; place[low]=place[j];  place[j]=k;
   if (j < mid) 
       median_find(j+1,high,mid);
   else if (j > mid) 
       median_find(low,j-1,mid);
   else
       mid_point = j;
}
   
void  b_level () {
  /* find out BLEVEL time for each node, O(n+e) */
  for (; readynum>0 ; ) {
    nid = ready[--readynum]; nodep = &(graph[nid]);
    for (ptr=nodep->parent; ptr!=NULL; ptr=ptr->linkNext) {
      pnodep = &(graph[ptr->nid]); 
      t = nodep->blevel + nodep->compcost;
      if ((t+ptr->commcost) > pnodep->blevel) {
	pnodep->blevel = t + ptr->commcost;
      }
      if (++pnodep->visit==pnodep->childnum) {
	ready[readynum++] = ptr->nid;
      }
    }
    nodep->blevel+=nodep->compcost;
  }
}

void  t_level () {
    /* find out TLEVEL time for each node, O(n+e) */
    for (; readynum2>0;) {
	nid=ready2[--readynum2]; nodep = &(graph[nid]);
	t = nodep->tlevel + nodep->compcost;
        for (ptr=nodep->child; ptr!=NULL; ptr=ptr->linkNext) {
 	    pnodep = &(graph[ptr->nid]); 
	    if ((t+ptr->commcost) > pnodep->tlevel)
	      pnodep->tlevel = t+ptr->commcost;
	    if (++pnodep->visit == pnodep->pnum) {
	      ready2[readynum2++] = pnodep->nid;    
	      pnodep->visit = 0;
	    }
	}
    }
}

void insert_list(int pos) 
{
  pri_list[++list_length] = pos;
}


int  get_highest_p_ready_node()
{
  int i, n_id, place;
  float t, earliest_finish_time, priority;
  priority = - 1.0 * MAXNUM;
  
  if (list_length >= 1) {
    for (place=1, i=1; i<=list_length; i++) {
      if  (graph[pri_list[i]].blevel + graph[pri_list[i]].tlevel  
	   > priority) {
	place = i;
	priority = graph[pri_list[i]].blevel + graph[pri_list[i]].tlevel;
      }
    }
    n_id = pri_list[place];
    pri_list[place] = pri_list[list_length];
    list_length--;
    return n_id;
  }
  else {
    printf("there is an error \n");
    exit(1);
  }
}
